package org.sat4j.tools;

import java.io.Serializable;

import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.TimeoutException;

/**
 * Converts Dimacs problems in array format (without the terminating 0)
 * to Dimacs Strings.
 * 
 * Adaptation of org.sat4j.reader.DimacsReader.
 * 
 * @author dlb
 * @author or
 * @author fuhs
 */
public class DimacsArrayToDimacsConverter implements Serializable {

    private static final long serialVersionUID = 1L;

    // counter for the number of clauses that occur in the SAT instance
    protected int clauses;
    
    protected StringBuilder dimacs; // stores the dimacs string while under construction
    private final int bufSize; // initial capacity of dimacs

    protected volatile boolean undertimeout; // have we received an expireTimeout()?

    // we check the volatile attribute this.undertimeout
    // each TIMEOUT_ITERATION_CHECKS-th iteration
    // in parseInstance(...) to keep the CPU cache happy
    // (possibly premature optimization)
    private static final int TIMEOUT_ITERATION_CHECKS = 100;

    public DimacsArrayToDimacsConverter(int bufSize) {
        this.bufSize = bufSize;
        this.undertimeout = true;
    }

    protected boolean handleConstr(int gateType, int output, int[] inputs)
            throws ContradictionException {
        for (int var : inputs) {
            this.dimacs.append(var);
            this.dimacs.append(" ");
        }
        this.dimacs.append("0\n");
        ++this.clauses;
        return true;
    }

    /**
     * @param gateType
     *             gateType[i] is the type of gate i according to the
     *  Extended Dimacs specs; ignored in DimacsArrayReader, but important
     *  for inheriting classes
     * @param outputs
     *             outputs[i] is the number of the output; ignored in
     *  DimacsArrayReader
     * @param inputs
     *             inputs[i] contains the clauses in DimacsArrayReader;
     *  an overriding class might have it contain the inputs of the current
     *  gate
     * @param maxVar
     *             the maximum number of assigned ids
     * @return a Dimacs String representation of the passed arguments
     *  computed using Tseitin's algorithm (does not require the input
     *  to be in NNF)
     * @throws ContradictionException
     *             si le probleme est trivialement inconsitant
     * @throws TimeoutException
     *             if a call to expireTimeout() was detected
     */
    public String parseInstance(int[] gateType,
            int[] outputs, int[][] inputs, int maxVar)
            throws ContradictionException, TimeoutException {
        init();
        this.dimacs.append(maxVar);
        this.dimacs.append(" ");
        
        // the first character to be replaced is saved
        // (7 = "p cnf ".length() + " ".length())
        int firstCharPos = 7 + Integer.toString(maxVar).length();
        
        this.dimacs.append("                    ");
        // 20 blanks; if the number of clauses ever exceeds 10^21-1, this needs
        // to be altered. But that would require BigIntegers anyway.
        
        this.dimacs.append("\n");
        for (int i = 0; i < outputs.length; ++i) {
            if (i % TIMEOUT_ITERATION_CHECKS == 0
                && (! this.undertimeout)) {
                throw new TimeoutException("Terminating parseInstance(...) as desired.");
            }
            handleConstr(gateType[i], outputs[i], inputs[i]);
        } 
        String numClauses = Integer.toString(this.clauses);
        int numClausesLength = numClauses.length();
        for (int i = 0; i < numClausesLength; ++i) {
            this.dimacs.setCharAt(firstCharPos + i, numClauses.charAt(i));
        }
        if (! this.undertimeout) {
            throw new TimeoutException("Terminating parseInstance(...) as desired (almost finished).");
        }
        String result = this.dimacs.toString();
        this.dimacs = null; // let the garbage collector at it
        if (! this.undertimeout) {
            throw new TimeoutException("Terminating parseInstance(...) as desired (so close!).");
        }
        return result;
    }

    protected void init() {
        this.dimacs = new StringBuilder(this.bufSize);
        this.dimacs.append("p cnf ");
        this.clauses = 0;
    }
    
    public String decode(int[] model) {
        StringBuilder stb = new StringBuilder();
        for (int i = 0; i < model.length; i++) {
            stb.append(model[i]);
            stb.append(" ");
        }
        stb.append("0");
        return stb.toString();
    }

    /**
     * Call this method to indicate that you do not care about the on-going
     * computations performed by this any more, but instead you now want
     * the currently executed method of this to terminate with an arbitrary
     * result.
     */
    public void expireTimeout() {
        this.undertimeout = false;
    }
}
