package org.sat4j.tools;

import java.io.Serializable;

import org.sat4j.specs.ContradictionException;

/**
 * Converts Dimacs problems in array format (without the terminating 0)
 * to 2 Dimacs Strings, one of them holds the reversed formula.
 * 
 * Adaptation of org.sat4j.reader.DimacsReader.
 * 
 * @author dlb
 * @author or
 * @author fuhs
 * @author Andreas Kelle-Emden
 */
public class DimacsArrayToDimacsReverseConverter implements Serializable {

    private static final long serialVersionUID = 1L;

    // counter for the number of clauses that occur in the SAT instance
    protected int clauses;
    protected int clausesReverse;
    
    protected StringBuilder dimacs; // stores the dimacs string while under construction
    protected StringBuilder dimacsReverse; // stores the dimacs string while under construction
    private final int bufSize; // initial capacity of dimacs

    public DimacsArrayToDimacsReverseConverter(int bufSize) {
        this.bufSize = bufSize;
    }

    protected boolean handleConstr(int gateType, int output, int[] inputs)
            throws ContradictionException {
        for (int var : inputs) {
            this.dimacs.append(var);
            this.dimacs.append(" ");
        }
        this.dimacs.append("0\n");
        ++this.clauses;
        return true;
    }

    protected boolean handleConstrReverse(int gateType, int output, int[] inputs)
            throws ContradictionException {
        for (int var : inputs) {
            this.dimacsReverse.append(var);
            this.dimacsReverse.append(" ");
        }
        this.dimacsReverse.append("0\n");
        ++this.clausesReverse;
        return true;
    }


    public static class Pair<S, T> {
        public S x;
        public T y;
        
        public Pair (S x, T y) {
            this.x = x;
            this.y = y;
        }
    }
    
    /**
     * @param gateType
     *             gateType[i] is the type of gate i according to the
     *  Extended Dimacs specs; ignored in DimacsArrayReader, but important
     *  for inheriting classes
     * @param outputs
     *             outputs[i] is the number of the output; ignored in
     *  DimacsArrayReader
     * @param inputs
     *             inputs[i] contains the clauses in DimacsArrayReader;
     *  an overriding class might have it contain the inputs of the current
     *  gate
     * @param maxVar
     *             the maximum number of assigned ids  
     * @throws ContradictionException
     *             si le probleme est trivialement inconsitant
     */
    public Pair<String, String> parseInstance(int[] gateType,
            int[] outputs, int[][] inputs, int maxVar)
            throws ContradictionException {
        init();
        this.dimacs.append(maxVar);
        this.dimacsReverse.append(maxVar);
        this.dimacs.append(" ");
        this.dimacsReverse.append(" ");
        
        // the first character to be replaced is saved
        // (7 = "p cnf ".length() + " ".length())
        int firstCharPos = 7 + Integer.toString(maxVar).length();
        
        this.dimacs.append("                    ");
        this.dimacsReverse.append("                    ");
        // 20 blanks; if the number of clauses ever exceeds 10^21-1, this needs
        // to be altered. But that would require BigIntegers anyway.
        
        this.dimacs.append("\n");
        this.dimacsReverse.append("\n");
        for (int i = 0; i < outputs.length; ++i) {
            handleConstr(gateType[i], outputs[i], inputs[i]); 
        } 
//        for (int i = 0; i < outputs.length; ++i) {
        for (int i = outputs.length-1; i >= 0; --i) {
            handleConstrReverse(gateType[i], outputs[i], inputs[i]); 
        } 
        String numClauses = Integer.toString(this.clauses);
        int numClausesLength = numClauses.length();
        for (int i = 0; i < numClausesLength; ++i) {
            this.dimacs.setCharAt(firstCharPos + i, numClauses.charAt(i));
        }
        String numClausesReverse = Integer.toString(this.clausesReverse);
        int numClausesLengthReverse = numClausesReverse.length();
        for (int i = 0; i < numClausesLengthReverse; ++i) {
            this.dimacsReverse.setCharAt(firstCharPos + i, numClausesReverse.charAt(i));
        }
        String result = this.dimacs.toString();
        this.dimacs = null; // let the garbage collector at it
        String resultReverse = this.dimacsReverse.toString();
        this.dimacsReverse = null; // let the garbage collector at it
        return new Pair<String, String> (result, resultReverse);
    }

    protected void init() {
        this.dimacs = new StringBuilder(this.bufSize);
        this.dimacs.append("p cnf ");
        this.dimacsReverse = new StringBuilder(this.bufSize);
        this.dimacsReverse.append("p cnf ");
        this.clauses = 0;
        this.clausesReverse = 0;
    }
    
    public String decode(int[] model) {
        StringBuilder stb = new StringBuilder();
        for (int i = 0; i < model.length; i++) {
            stb.append(model[i]);
            stb.append(" ");
        }
        stb.append("0");
        return stb.toString();
    }
}
