/*
 * SAT4J: a SATisfiability library for Java Copyright (C) 2004-2006 Daniel Le
 * Berre
 * 
 * Based on the original minisat specification from:
 * 
 * An extensible SAT solver. Niklas E?n and Niklas S?rensson. Proceedings of the
 * Sixth International Conference on Theory and Applications of Satisfiability
 * Testing, LNCS 2919, pp 502-518, 2003.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

package org.sat4j.tools;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Map;

import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.IConstr;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.IVec;
import org.sat4j.specs.IVecInt;
import org.sat4j.specs.TimeoutException;

/**
 * Solver used to write down a CNF into a String.
 * 
 * It is especially useful compared to the DimacsOutputSolver 
 * because the number of clauses does not need to be known in 
 * advance.
 * 
 * Backport from developer version in November 2007.
 * 
 * @author leberre
 */
public class DimacsStringSolver implements ISolver {

    /**
     * 
     */
    private static final long serialVersionUID = 1L;

    private transient StringBuilder out;

    private int nbvars;

    private int nbclauses;

    private boolean fixedNbClauses = false;

    private boolean firstConstr = true;

    private int firstCharPos;
    
    private final int initBuilderSize;
    
    public DimacsStringSolver() {
        this(16);
    }

    public DimacsStringSolver(int initSize) {
        out = new StringBuilder(initSize);
        initBuilderSize = initSize;
    }


    public int newVar() {
        return 0;
    }

    public int newVar(int howmany) {
        out.append("p cnf " + howmany);
        nbvars = howmany;
       
        return howmany;
    }

    public void setExpectedNumberOfClauses(int nb) {
        out.append(" " + nb);
        nbclauses = nb;
        fixedNbClauses = true;
    }

    public IConstr addClause(IVecInt literals) throws ContradictionException {
        if (firstConstr) {
            if (!fixedNbClauses) {
                firstCharPos = 7 + Integer.toString(nbvars).length();
                out.append("                    ");
                out.append("\n");
                nbclauses=0;
            }
            firstConstr = false;
        }
        if (!fixedNbClauses) nbclauses++;
        
        for (int i : literals)
            out.append(i).append(" ");
        out.append("0\n");
        return null;        
    }

    public boolean removeConstr(IConstr c) {
        throw new UnsupportedOperationException();
    }

    public void addAllClauses(IVec<IVecInt> clauses)
            throws ContradictionException {
        throw new UnsupportedOperationException();
    }

    public IConstr addAtMost(IVecInt literals, int degree)
            throws ContradictionException {
        if (degree > 1) {
            throw new UnsupportedOperationException(
                    "Not a clausal problem! degree " + degree);
        }
        assert degree == 1;
        if (firstConstr) {
            if (!fixedNbClauses) {
                firstCharPos = 7 + Integer.toString(nbvars).length();
                out.append("                    ");
                out.append("\n");
                nbclauses=0;
            }
            firstConstr = false;
        }
        
        for (int i = 0; i <= literals.size(); i++) {
            for (int j = i + 1; j < literals.size(); j++) {
                if (!fixedNbClauses) nbclauses++;
                out.append("" + (-literals.get(i)) + " " + (-literals.get(j))
                        + " 0\n");
            }
        }
        return null;
    }

    public IConstr addAtLeast(IVecInt literals, int degree)
            throws ContradictionException {
        if (degree > 1) {
            throw new UnsupportedOperationException(
                    "Not a clausal problem! degree " + degree);
        }
        assert degree == 1;
        return addClause(literals);
    }

    public IConstr addPseudoBoolean(IVecInt lits, IVec<BigInteger> coeffs,
            boolean moreThan, BigInteger d) throws ContradictionException {
        throw new UnsupportedOperationException();
    }

    public void setTimeout(int t) {
        // TODO Auto-generated method stub

    }

    public void setTimeoutMs(long t) {
        // TODO Auto-generated method stub
    }

    public int getTimeout() {
        // TODO Auto-generated method stub
        return 0;
    }

    public void reset() {
        fixedNbClauses = false;
        firstConstr  = true;
        out = new StringBuilder(initBuilderSize);
    }

    public void printStat(PrintStream out, String prefix) {
        // TODO Auto-generated method stub

    }

    public void printStat(PrintWriter out, String prefix) {
        // TODO Auto-generated method stub

    }

    public Map<String, Number> getStat() {
        // TODO Auto-generated method stub
        return null;
    }

    public String toString(String prefix) {
        return "Dimacs output solver";
    }

    public void clearLearntClauses() {
        // TODO Auto-generated method stub

    }

    public int[] model() {
        throw new UnsupportedOperationException();
    }

    public boolean model(int var) {
        throw new UnsupportedOperationException();
    }

    public boolean isSatisfiable() throws TimeoutException {
        throw new TimeoutException("There is no real solver behind!");
    }

    public boolean isSatisfiable(IVecInt assumps) throws TimeoutException {
        throw new TimeoutException("There is no real solver behind!");
    }

    public int[] findModel() throws TimeoutException {
        throw new UnsupportedOperationException();
    }

    public int[] findModel(IVecInt assumps) throws TimeoutException {
        throw new UnsupportedOperationException();
    }

    public int nConstraints() {
        return nbclauses;
    }

    public int nVars() {
        return nbvars;
    }
    
    @Override
    public String toString() {
        String numClauses = Integer.toString(nbclauses);
        int numClausesLength = numClauses.length();
        for (int i = 0; i < numClausesLength; ++i) {
            out.setCharAt(firstCharPos + i, numClauses.charAt(i));
        }
        return out.toString();
    }



    public void expireTimeout() {
        // TODO Auto-generated method stub
        
    }
}
