package org.sat4j.tools;

import org.sat4j.core.VecInt;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.IVecInt;

/**
 * Converter from the Extended Dimacs format proposed by Fahiem Bacchus 
 * and Toby Walsh in array representation (without the terminating 0)
 * to the Dimacs format.
 * 
 * Adaptation of org.sat4j.reader.ExtendedDimacsReader.
 * 
 * @author leberre
 * @author fuhs
 */
public class ExtendedDimacsArrayToDimacsConverter extends DimacsArrayToDimacsConverter {

    public static final int FALSE = 1;

    public static final int TRUE = 2;

    public static final int NOT = 3;

    public static final int AND = 4;

    public static final int NAND = 5;

    public static final int OR = 6;

    public static final int NOR = 7;

    public static final int XOR = 8;

    public static final int XNOR = 9;

    public static final int IMPLIES = 10;

    public static final int IFF = 11;

    public static final int IFTHENELSE = 12;

    public static final int ATLEAST = 13;

    public static final int ATMOST = 14;

    public static final int COUNT = 15;

    private static final long serialVersionUID = 1L;

    // experimental code for XOR solving with Erika Abraham 
    private static final int NORMAL_CLAUSE_PREFIX = 1; 
    private static final int XOR_CLAUSE_PREFIX = -1; 
    
    private final boolean useXorClauses;
    // end experimental code for XOR solving with Erika Abraham

    public ExtendedDimacsArrayToDimacsConverter(int bufSize) {
        super(bufSize);
        this.useXorClauses = false;
    }
    
    public ExtendedDimacsArrayToDimacsConverter(int bufSize, boolean useXorClauses) {
        super(bufSize);
        this.useXorClauses = useXorClauses;
    }

    /**
     * Handles a single constraint (constraint == Extended Dimacs circuit
     * gate).
     * 
     * @param gateType
     *             the type of the gate in question
     * @param output
     *             the number of the output of the gate in question
     * @param inputs
     *             the numbers of the inputs of the gates in question
     * @return true
     */
    @Override
    protected boolean handleConstr(int gateType, int output, int[] inputs)
            throws ContradictionException {
        IVecInt literals = new VecInt(inputs);
        switch (gateType) {
            case FALSE:
                gateFalse(output, literals);
                break;
            case TRUE:
                gateTrue(output, literals);
                break;
            case OR:
                or(output, literals);
                break;
            case NOT:
                not(output, literals);
                break;
            case AND:
                and(output, literals);
                break;
            case XOR:
                xor(output, literals);
                break;
            case IFF:
                iff(output, literals);
                break;
            case IFTHENELSE:
                ite(output, literals);
                break;
            default:
                throw new UnsupportedOperationException("Gate type "
                        + gateType + " not handled yet");
        }
        return true;
    }

    private void gateFalse(int y, IVecInt literals)
            throws ContradictionException {
        assert literals.size() == 0;
        IVecInt clause = new VecInt(1);
        clause.push(-y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
    }

    private void gateTrue(int y, IVecInt literals)
            throws ContradictionException {
        assert literals.size() == 0;
        IVecInt clause = new VecInt(1);
        clause.push(y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
    }

    private void ite(int y, IVecInt literals) throws ContradictionException {
        assert literals.size() == 3;
        IVecInt clause = new VecInt(3);
        // y <=> (x1 -> x2) and (not x1 -> x3)
        final int x1, x2, x3;
        x1 = literals.get(0);
        x2 = literals.get(1);
        x3 = literals.get(2);
        // y -> (x1 -> x2) and (not x1 -> x3)
        clause.push(-y).push(-x1).push(x2);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        clause.clear();
        clause.push(-y).push(x1).push(x3);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        // y <- (x1 -> x2) and (not x1 -> x3)
        // not(x1 -> x2) or not(not x1 -> x3) or y
        // x1 and not x2 or not x1 and not x3 or y
        // (x1 and not x2) or ((not x1 or y) and (not x3 or y))
        // (x1 or not x1 or y) and (not x2 or not x1 or y) and (x1 or not x3 or
        // y) and (not x2 or not x3 or y)
        // not x1 or not x2 or y and x1 or not x3 or y and not x2 or not x3 or y
        clause.clear();
        clause.push(-x1).push(-x2).push(y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        clause.clear();
        clause.push(x1).push(-x3).push(y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        clause.clear();
        clause.push(-x2).push(-x3).push(y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
    }

    private void and(int y, IVecInt literals) throws ContradictionException {
        // y <=> AND x1 ... xn
        final int length = literals.size();

        // y <= x1 .. xn
        IVecInt clause = new VecInt(length + 1);
        clause.push(y);
        for (int i = 0; i < length; i++) {
            clause.push(-literals.get(i));
        }
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        clause.clear();
        for (int i = 0; i < length; i++) {
            // y => xi
            clause.clear();
            clause.push(-y);
            clause.push(literals.get(i));
            processClause(clause, NORMAL_CLAUSE_PREFIX);
        }
    }

    private void or(int y, IVecInt literals) throws ContradictionException {
        // y <=> OR x1 x2 ...xn
        final int length = literals.size();

        // y => x1 x2 ... xn
        IVecInt clause = new VecInt(length + 1);
        literals.copyTo(clause);
        clause.push(-y);
        processClause(clause, NORMAL_CLAUSE_PREFIX);
        clause.clear();
        for (int i = 0; i < length; i++) {
            // xi => y
            clause.clear();
            clause.push(y);
            clause.push(-literals.get(i));
            processClause(clause, NORMAL_CLAUSE_PREFIX);
        }
    }

    private void processClause(IVecInt clause, int clausePrefix) throws ContradictionException {
        if (this.useXorClauses) {
            this.dimacs.append(clausePrefix);
            this.dimacs.append(" ");
        }
        final int length = clause.size();
        for (int i = 0; i < length; ++i) {
            this.dimacs.append(clause.get(i));
            this.dimacs.append(" ");
        }
        this.dimacs.append("0\n");
        ++this.clauses;
    }

    private void not(int y, IVecInt literals) throws ContradictionException {
        assert literals.size() == 1;
        IVecInt clause = new VecInt(2);
        if (this.useXorClauses) {
            // y <-> not x  is equivalent to  y xor x
            clause.push(y).push(literals.get(0));
            processClause(clause, XOR_CLAUSE_PREFIX);
        }
        else {
            // y <=> not x
            // y => not x = not y or not x
            final int x = literals.get(0);
            clause.push(-y).push(-x);
            processClause(clause, NORMAL_CLAUSE_PREFIX);
            // y <= not x = y or x
            clause.clear();
            clause.push(y).push(x);
            processClause(clause, NORMAL_CLAUSE_PREFIX);
        }
    }

    void xor(int y, IVecInt literals) throws ContradictionException {
        if (this.useXorClauses) {
            final int length = literals.size();
            IVecInt xorClause = new VecInt(1+length);
            xorClause.push(-y);
            for (int i = 0; i < length; ++i) {
                xorClause.push(literals.get(i));
            }
            processClause(xorClause, XOR_CLAUSE_PREFIX);
        }
        else {
            literals.push(-y);
            int[] f = new int[literals.size()];
            literals.copyTo(f);
            xor2Clause(f, 0, false);
        }
    }

    void iff(int y, IVecInt literals) throws ContradictionException {
        if (this.useXorClauses) {
            // analogous to xor:  y <-> (x1 <-> ... <-> xn)
            // is equivalent to  -y xor x1 xor ... xor xn  if n is odd
            //                    y xor x1 xor ... xor xn  if n is even
            final int length = literals.size();
            IVecInt xorClause = new VecInt(1+length);
            xorClause.push(length % 2 == 0 ? y : -y);
            for (int i = 0; i < length; ++i) {
                xorClause.push(literals.get(i));
            }
            processClause(xorClause, XOR_CLAUSE_PREFIX);
            
        }
        else {
            literals.push(y);
            int[] f = new int[literals.size()];
            literals.copyTo(f);
            iff2Clause(f, 0, false);
        }
    }

    void xor2Clause(int[] f, int prefix, boolean negation)
            throws ContradictionException {
        if (prefix == f.length - 1) {
            IVecInt clause = new VecInt(f.length);
            for (int i = 0; i < f.length - 1; ++i) {
                clause.push(f[i]);
            }
            clause.push(f[f.length - 1] * (negation ? -1 : 1));
            processClause(clause, NORMAL_CLAUSE_PREFIX);
            return;
        }

        if (negation) {
            f[prefix] = -f[prefix];
            xor2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];

            xor2Clause(f, prefix + 1, true);
        } else {
            xor2Clause(f, prefix + 1, false);

            f[prefix] = -f[prefix];
            xor2Clause(f, prefix + 1, true);
            f[prefix] = -f[prefix];
        }
    }

    void iff2Clause(int[] f, int prefix, boolean negation)
            throws ContradictionException {
        if (prefix == f.length - 1) {
            IVecInt clause = new VecInt(f.length);
            for (int i = 0; i < f.length - 1; ++i) {
                clause.push(f[i]);
            }
            clause.push(f[f.length - 1] * (negation ? -1 : 1));
            processClause(clause, NORMAL_CLAUSE_PREFIX);
            return;
        }

        if (negation) {
            iff2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, true);
            f[prefix] = -f[prefix];            
        } else {
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, true);
        }
    }

}
