package org.sat4j.tools;

import java.io.Serializable;

import org.sat4j.core.*;
import org.sat4j.specs.*;

/**
 * Create Partial Max-SAT input from Dimacs array.
 * 
 * SpecialAnd tokens will get weight 1, all other tokens will get weight TOP.
 * 
 * Adaptation of org.sat4j.tools.ExtendedDimacsArrayToDimacsConverter
 * 
 * @author dlb
 * @author or
 * @author fuhs
 * @author Andreas Kelle-Emden
 */
public class PMaxDimacsConverter implements Serializable {

    private static final long serialVersionUID = 1L;

    private static int TOP = Integer.MAX_VALUE - 1;
    private int weight = 0;
    
    // counter for the number of clauses that occur in the SAT instance
    protected int clauses;
    
    protected StringBuilder dimacs; // stores the dimacs string while under construction
    protected StringBuilder dimacs2; // stores the dimacs string while under construction
    private final int bufSize; // initial capacity of dimacs

    public PMaxDimacsConverter(int bufSize) {
        this.bufSize = bufSize;
    }
    public static final int FALSE = 1;

    public static final int TRUE = 2;

    public static final int NOT = 3;

    public static final int AND = 4;

    public static final int NAND = 5;

    public static final int OR = 6;

    public static final int NOR = 7;

    public static final int XOR = 8;

    public static final int XNOR = 9;

    public static final int IMPLIES = 10;

    public static final int IFF = 11;

    public static final int IFTHENELSE = 12;

    public static final int ATLEAST = 13;

    public static final int ATMOST = 14;

    public static final int COUNT = 15;

    /**
     * Handles a single constraint (constraint == Extended Dimacs circuit
     * gate).
     * 
     * @param gateType
     *             the type of the gate in question
     * @param output
     *             the number of the output of the gate in question
     * @param inputs
     *             the numbers of the inputs of the gates in question
     * @return true
     */
    protected boolean handleConstr(int gateType, int output, int[] inputs)
            throws ContradictionException {
        //this.dimacs2.append(this.weight+" ");
        IVecInt literals = new VecInt(inputs);
        switch (gateType) {
            case FALSE:
                gateFalse(output, literals);
                break;
            case TRUE:
                gateTrue(output, literals);
                break;
            case OR:
                or(output, literals);
                break;
            case NOT:
                not(output, literals);
                break;
            case AND:
                and(false, output, literals);
                break;
            case XOR:
                xor(output, literals);
                break;
            case IFF:
                iff(output, literals);
                break;
            case IFTHENELSE:
                ite(output, literals);
                break;
            default:
                throw new UnsupportedOperationException("Gate type "
                        + gateType + " not handled yet");
        }
        return true;
    }

    private void gateFalse(int y, IVecInt literals)
            throws ContradictionException {
        assert literals.size() == 0;
        IVecInt clause = new VecInt(1);
        clause.push(-y);
        processClause(clause);
    }

    private void gateTrue(int y, IVecInt literals)
            throws ContradictionException {
        assert literals.size() == 0;
        IVecInt clause = new VecInt(1);
        clause.push(y);
        processClause(clause);
    }

    private void ite(int y, IVecInt literals) throws ContradictionException {
        assert literals.size() == 3;
        IVecInt clause = new VecInt(2);
        // y <=> (x1 -> x2) and (not x1 -> x3)
        // y -> (x1 -> x2) and (not x1 -> x3)
        clause.push(-y).push(-literals.get(0)).push(literals.get(1));
        processClause(clause);
        clause.clear();
        clause.push(-y).push(literals.get(0)).push(literals.get(2));
        processClause(clause);
        // y <- (x1 -> x2) and (not x1 -> x3)
        // not(x1 -> x2) or not(not x1 -> x3) or y
        // x1 and not x2 or not x1 and not x3 or y
        // (x1 and not x2) or ((not x1 or y) and (not x3 or y))
        // (x1 or not x1 or y) and (not x2 or not x1 or y) and (x1 or not x3 or
        // y) and (not x2 or not x3 or y)
        // not x1 or not x2 or y and x1 or not x3 or y and not x2 or not x3 or y
        clause.clear();
        clause.push(-literals.get(0)).push(-literals.get(1)).push(y);
        processClause(clause);
        clause.clear();
        clause.push(literals.get(0)).push(-literals.get(2)).push(y);
        processClause(clause);
        clause.clear();
        clause.push(-literals.get(1)).push(-literals.get(2)).push(y);
        processClause(clause);
    }

    private void and(boolean isPMaxSatAnd, int y, IVecInt literals) throws ContradictionException {
        // y <=> AND x1 ... xn

        // y <= x1 .. xn
        IVecInt clause = new VecInt(literals.size() + 1);
        clause.push(y);
        for (int i = 0; i < literals.size(); i++) {
            clause.push(-literals.get(i));
        }
        processClause(clause);
        clause.clear();
        for (int i = 0; i < literals.size(); i++) {
            // y => xi
            clause.clear();
            clause.push(-y);
            clause.push(literals.get(i));
            processClause(clause);
        }
    }

    private void or(int y, IVecInt literals) throws ContradictionException {
        // y <=> OR x1 x2 ...xn
        // y => x1 x2 ... xn
        IVecInt clause = new VecInt(literals.size() + 1);
        literals.copyTo(clause);
        clause.push(-y);
        processClause(clause);
        clause.clear();
        for (int i = 0; i < literals.size(); i++) {
            // xi => y
            clause.clear();
            clause.push(y);
            clause.push(-literals.get(i));
            processClause(clause);
        }
    }

    private void processClause(IVecInt clause) throws ContradictionException {
        final int length = clause.size();
        this.dimacs2.append(this.weight + " ");
        for (int i = 0; i < length; ++i) {
            this.dimacs2.append(clause.get(i));
            this.dimacs2.append(" ");
        }
        this.dimacs2.append("0\n");
        ++this.clauses;
    }

    private void not(int y, IVecInt literals) throws ContradictionException {
        assert literals.size() == 1;
        IVecInt clause = new VecInt(2);
        // y <=> not x
        // y => not x = not y or not x
        clause.push(-y).push(-literals.get(0));
        processClause(clause);
        // y <= not x = y or x
        clause.clear();
        clause.push(y).push(literals.get(0));
        processClause(clause);
    }

    void xor(int y, IVecInt literals) throws ContradictionException {
        literals.push(-y);
        int[] f = new int[literals.size()];
        literals.copyTo(f);
        xor2Clause(f, 0, false);
    }

    void iff(int y, IVecInt literals) throws ContradictionException {
        literals.push(y);
        int[] f = new int[literals.size()];
        literals.copyTo(f);
        iff2Clause(f, 0, false);
    }

    void xor2Clause(int[] f, int prefix, boolean negation)
            throws ContradictionException {
        if (prefix == f.length - 1) {
            IVecInt clause = new VecInt(f.length);
            for (int i = 0; i < f.length - 1; ++i) {
                clause.push(f[i]);
            }
            clause.push(f[f.length - 1] * (negation ? -1 : 1));
            processClause(clause);
            return;
        }

        if (negation) {
            f[prefix] = -f[prefix];
            xor2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];

            xor2Clause(f, prefix + 1, true);
        } else {
            xor2Clause(f, prefix + 1, false);

            f[prefix] = -f[prefix];
            xor2Clause(f, prefix + 1, true);
            f[prefix] = -f[prefix];
        }
    }

    void iff2Clause(int[] f, int prefix, boolean negation)
            throws ContradictionException {
        if (prefix == f.length - 1) {
            IVecInt clause = new VecInt(f.length);
            for (int i = 0; i < f.length - 1; ++i) {
                clause.push(f[i]);
            }
            clause.push(f[f.length - 1] * (negation ? -1 : 1));
            processClause(clause);
            return;
        }

        if (negation) {
            iff2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, true);
            f[prefix] = -f[prefix];            
        } else {
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, false);
            f[prefix] = -f[prefix];
            iff2Clause(f, prefix + 1, true);
        }
    }

/*    protected boolean handleConstr(int weight, int gateType, int output, int[] inputs)
            throws ContradictionException {
        this.dimacs2.append(weight+" ");
        for (int var : inputs) {
            this.dimacs2.append(var);
            this.dimacs2.append(" ");
        }
        this.dimacs2.append("0\n");
        ++this.clauses;
        return true;
    }
*/
    
    /**
     * @param gateType
     *             gateType[i] is the type of gate i according to the
     *  Extended Dimacs specs; ignored in DimacsArrayReader, but important
     *  for inheriting classes
     * @param outputs
     *             outputs[i] is the number of the output; ignored in
     *  DimacsArrayReader
     * @param inputs
     *             inputs[i] contains the clauses in DimacsArrayReader;
     *  an overriding class might have it contain the inputs of the current
     *  gate
     * @param maxVar
     *             the maximum number of assigned ids  
     * @throws ContradictionException
     *             si le probleme est trivialement inconsitant
     */
    public String parseInstance(
            int[] gateType, int[] outputs, int[][] inputs,
            int maxVar)
            throws ContradictionException {
        init();
        this.dimacs.append(maxVar);
        this.dimacs.append(" ");
        
        this.weight = TOP;
        for (int i = 0; i < outputs.length; ++i) {
            handleConstr(gateType[i], outputs[i], inputs[i]); 
        } 
        this.dimacs.append(this.clauses + " " + TOP + "\n");
        this.dimacs.append(this.dimacs2);
        String result = this.dimacs.toString();
        this.dimacs = null; // let the garbage collector at it
        this.dimacs2 = null; // let the garbage collector at it
        return result;
    }

    protected void init() {
        this.dimacs = new StringBuilder(this.bufSize);
        this.dimacs2 = new StringBuilder(this.bufSize);
        this.dimacs.append("p wcnf ");
        this.clauses = 0;
    }
    
    public String decode(int[] model) {
        StringBuilder stb = new StringBuilder();
        for (int i = 0; i < model.length; i++) {
            stb.append(model[i]);
            stb.append(" ");
        }
        stb.append("0");
        return stb.toString();
    }
}